#!/usr/bin/env python

from svm import *
import math, string, random
import os, sys

gnuplot_exe="/usr/bin/gnuplot"

#Sigmoid_Training(deci[],label[],prior1,prior0)
#input decision values, real_label{1,-1}, #positive instances, #negative instances
#output A,B that minimize sigmoid likilihood
def Sigmoid_Training(deci,label,prior1,prior0):
#Parameter Setting
	maxiter=100	#Maximum number of iterations
	minstep=1e-10	#Minimum step taken in line search
	sigma=1e-3	#For numerically strict PD of Hessian
	eps=1e-5
	
#Construct Target Support
	hiTarget=(prior1+1.0)/(prior1+2.0)
	loTarget=1/(prior0+2.0)
	len=prior1+prior0
	t=label[:]  #simply want a len sized vector
	for i in range(len):
		if label[i] > 0:
			t[i] = hiTarget
		else:
			t[i] = loTarget

#Initial Point and Initial Fun Value
	A,B,fval=0.0, math.log((prior0+1.0)/(prior1+1.0)), 0.0
	for i in range(len):
		fApB = deci[i]*A+B
		if fApB >= 0:
			fval += t[i]*fApB + math.log(1+math.exp(-fApB))
		else:
			fval += (t[i] - 1)*fApB +math.log(1+math.exp(fApB))

	for it in range(maxiter):
		#Update Gradient and Hessian (use H' = H + sigma I)
		h11=h22=sigma #Numerically ensures strict PD
		h21=g1=g2=0.0
		for i in range(len):
			fApB = deci[i]*A+B
			if (fApB >= 0):
				p=math.exp(-fApB)/(1.0+math.exp(-fApB))
				q=1.0/(1.0+math.exp(-fApB))
			else:
				p=1.0/(1.0+math.exp(fApB))
				q=math.exp(fApB)/(1.0+math.exp(fApB))
			d2=p*q
			h11+=deci[i]*deci[i]*d2
			h22+=d2
			h21+=deci[i]*d2
			d1=t[i]-p
			g1+=deci[i]*d1
			g2+=d1

		#Stopping Criteria
		if abs(g1)<eps and abs(g2)<eps:
			break

		#Finding Newton direction: -inv(H') * g
		det=h11*h22-h21*h21
		dA=-(h22*g1 - h21 * g2) / det
		dB=-(-h21*g1+ h11 * g2) / det
		gd=g1*dA+g2*dB

		#Line Search
		stepsize = 1
		while stepsize >= minstep:
			newA = A + stepsize * dA
			newB = B + stepsize * dB

			#New function value
			newf = 0.0
			for i in range(len):
				fApB = deci[i]*newA+newB
				if fApB >= 0:
					newf += t[i]*fApB + math.log(1+math.exp(-fApB))
				else:
					newf += (t[i] - 1)*fApB +math.log(1+math.exp(fApB))

			#Check sufficient decrease
			if newf < fval + 0.0001 * stepsize * gd:
				A, B, fval = newA, newB, newf
				break
			else:
				stepsize = stepsize / 2.0

		if stepsize < minstep:
			print "line search fails",A,B,g1,g2,dA,dB,gd
			return [A,B]

	if it>=maxiter-1:
		print "reaching maximal iterations",g1,g2
	return [A,B]


#cross(prob_x[], prob_y[], svm_parameter param, nr_fold)
#input raw attributes, labels, param, cv_fold in decision value building
#output list of decision value
def cross(prob_x, prob_y, param, nr_fold):
	deci = []	#output of decision value
	prob_l = len(prob_y)


	#random permutation by swapping i and j instance
	for i in range(prob_l):
		j = random.randrange(i,prob_l)
		prob_x[i], prob_x[j] = prob_x[j], prob_x[i]
		prob_y[i], prob_y[j] = prob_y[j], prob_y[i]

	#cross training
	for i in range(nr_fold):
		begin = i * prob_l / nr_fold
		end = (i + 1) * prob_l / nr_fold
		subprob = svm_problem(prob_y[:begin] + prob_y[end:], prob_x[:begin] + prob_x[end:])
		submodel = svm_model(subprob, param)
		#predict and grab decision value
		for j in range(begin, end):
			deci.append(submodel.predict(prob_x[j]))
	
	return deci
#read svm sparse format from file 'fname'
#output: [attributes, labels, postive, negative]
def read_svm(dtinfo):
	attrs = []
	labels = []
	pos = 0
	neg = 0
	f = open(dtinfo[0], "r")
	line = f.readline()
	while line:
		elems = string.split(line)
		attr = {}
		for e in elems[1:]:
			points = string.split(e, ":")
			attr[int(points[0])] = float(points[1])
		attrs.append(attr)

		label = float(elems[0])
		labels.append(label)
		if label > 0:
			pos += 1
		else:
			neg += 1

		line = f.readline()
	f.close()
	return [attrs, labels, pos, neg]

#predpr(deci[], A, B)
#	return [[deci,prob]]
#reads lists of decision values and prediction mapping function parameter A,B
#then outputs list of [decision value, predicted probability]
#predpr(deci[], A, B)
#	return [[deci,prob]]
#reads lists of decision values and prediction mapping function parameter A,B
#then outputs list of [decision value, predicted probability]
def predpr(deci, A, B):
	if type(deci) == list:
		prob = []
		for i in range(len(deci)):
			fApB = deci[i] * A +B
			if (fApB >= 0):
				prob.append([deci[i], math.exp(-fApB)/(1.0+math.exp(-fApB))])
			else:
				prob.append([deci[i], 1.0/(1+math.exp(fApB)) ] ) 
		return prob
	else:
		if (deci*A+B >= 0):
			return math.exp(-(deci*A+B))/(1.0+math.exp(-(deci*A+B)))
		else:
			return 1.0/(1+math.exp(A*deci + B))

#realpr(deci[] label[])
#	return [[deci,pr]]
#
#reads lists of decision values and their real label
#then outputs list of [decision value, real Baysian probability]
def realpr(deci , label):
   	hispos = {}
   	hisneg = {}
	factor = 10.0
	for i in range(len(deci)):
   			hispos[int(deci[i]*factor)] = 0.0
   			hisneg[int(deci[i]*factor)] = 0.0

	for i in range(len(deci)):
		if label[i] > 0:
   			hispos[int(deci[i]*factor)] += 1.0
		else:
   			hisneg[int(deci[i]*factor)] += 1.0

	prob = {}
	for i in range(len(deci)):
   		index = int(deci[i]*factor) 
   		prpos = hispos[index]
   		prneg = hisneg[index]
   		prob[index/factor] = prpos / (prpos + prneg)

   	prob_arr = []
	for i in prob.keys():
		prob_arr.append([i, prob[i]])
	return prob_arr
#initialize gnuplot, 
#if term != x11 then a postscript file named '$term' is created
def plot_init(term='x11'):
	global gnuplot, gnuplot_exe

	gnuplot = os.popen(gnuplot_exe,'w')
	if term=='x11':
		gnuplot.write("set term x11\n")
	else:
		#gnuplot.write("set term postscript eps 22\n")
		gnuplot.write("set term png\n");
		#gnuplot.write("set output \"%s.ps\"\n" % (term))
		gnuplot.write("set output \"%s.png\"\n" % (term))
	gnuplot.write("set multiplot\n")

#terminate gnuplot
def plot_final():
	gnuplot.write("quit\n")
	gnuplot.flush()
	gnuplot.close()


#plot Data[[x,y]] with Lines? and Linewidth
def plot(Data, Line=0, width=7):
	if Line==0:
		gnuplot.write("plot \"-\" notitle pt 2 \n")
	else:
		gnuplot.write("plot \"-\" notitle with lines linewidth %d\n" % width)

	for i in range(len(Data)):
		gnuplot.write("%f %f\n" % (Data[i][0], Data[i][1]))
	gnuplot.write("e\n")
	gnuplot.flush()

#processing argv and set some global variables
def proc_argv(argv = sys.argv):
    #default parameter
	fold = 5
	cost = 2**0
	kern = 2
	gma = 0
	i = 1
	while i < len(argv) - 1:
		if argv[i] == '-v':
			i += 1
			fold = int(argv[i])
		elif argv[i] == '-g':
			i += 1
			gma = string.atof(argv[i])
		elif argv[i] == '-c':
			i += 1
			cost = string.atof(argv[i])
		elif argv[i] == '-t':
			i += 1
			kern = string.atoi(argv[i])
		i = i + 1

   	if gma!=0:
	   param = svm_parameter(svm_type=C_SVC, kernel_type=kern, gamma=gma, C=cost)
	else:
	   param = svm_parameter(svm_type=C_SVC, kernel_type=kern, C=cost)
	
	return param,fold

def proc_data(arg):
	dataset_pathname = arg
	dataset_title = os.path.split(dataset_pathname)[1]
	return [dataset_pathname,dataset_title]

def plot_roc(dtinfo,deci, label, A,B, pos, neg):
	plot_init(dtinfo[0]+'-roc')
	gnuplot.write("set xlabel \"False Positive Rate\"\n")
	gnuplot.write("set ylabel \"Ture Positive Rate\"\n")
	gnuplot.write("set title \"ROC curve of %s\"\n" % dtinfo[1])
	db = []
	prob = predpr(deci,A,B)
	for i in range(len(deci)):
		db.append([prob[i][1], label[i]])

	def cmp (x,y):
		if x[0] < y[0]:
			return 1
		else:
			return -1;

	db.sort(cmp)
	xy_arr = []
	tp=0.0
	fp=0.0
	for i in range(len(db)):
		if db[i][1]>0:		#positive
			tp+=1.0
		else:
			fp+=1.0
		xy_arr.append([fp/neg,tp/pos])

	plot(xy_arr, 1)
	f=open("%s-roc.txt" % dtinfo[0],"w")
	for i in range(len(xy_arr)):
		f.write("%f %f\n" % (xy_arr[i][0], xy_arr[i][1]))
	f.close()
   	plot_final()

def plot_prob(dtinfo,deci, label,A,B):
	plot_init(dtinfo[0]+'-prob')
	gnuplot.write("set xlabel \"Decision Value\"\n")
	gnuplot.write("set ylabel \"Predicted Probability\"\n")
	gnuplot.write("set title \"Probability v.s. Decision Value of %s\"\n" % dtinfo[1])
	splt= predpr([0],A,B)[0][1] 
	gnuplot.write("set label \"%.3f\" at %f,%f\n" % (splt, 0.3, splt))
   	#calculate margins
	factor = 2
	deci_min, deci_max = min(deci), max(deci)
	deci_mid = (deci_min + deci_max)/2
	deci_min = deci_mid + factor*(deci_min - deci_mid)
	deci_max = deci_mid + factor*(deci_max - deci_mid)
	gnuplot.write("set label \"A:%.3e\\nB:%.3e\" at %f,0.9\n" % (A,B, deci_min+0.1)) 
	x = deci_min
   	x_arr = []
	step = (deci_max - deci_min)/100.0
	while x<deci_max:
		x_arr.append(x)
   		x += step
	gnuplot.write("set xrange [%f:%f]\n" % (deci_min, deci_max))
	gnuplot.write("set yrange [%f:%f]\n" % (0.0,1.0))
	plot(realpr(deci,label), 0)
	gnuplot.write("set xrange [%f:%f]\n" % (deci_min, deci_max))
	gnuplot.write("set yrange [%f:%f]\n" % (0.0,1.0))
	f=open("%s-prob.txt" % dtinfo[0], "w")
	db = predpr(x_arr,A,B)
	for i in range(len(db)):
		f.write("%f %f\n" % (db[i][0], db[i][1]))
	f.close()
	plot(db, 1)
	plot([[0,0],[0,1]],1,1) 
	plot_final()

